@@ -1,4 +1,4 @@
-/*	$Id: lib.c,v 1.14 2017/06/24 14:38:32 schwarze Exp $ */
+/*	$Id: lib.c,v 1.15 2018/12/13 11:55:46 schwarze Exp $ */
 /*
  * Copyright (c) 2009 Kristaps Dzonsons <kristaps@bsd.lv>
  *
@@ -17,12 +17,9 @@
 #include "config.h"
 
 #include <sys/types.h>
-
 #include <string.h>
 
-#include "mandoc.h"
 #include "roff.h"
-#include "mdoc.h"
 #include "libmdoc.h"
 
 #define LINE(x, y) \
